#include <stdlib.h>
#include <utility>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <ctype.h>
#include <string.h>

#include "peddata.h"

using namespace std;


void kill_IQLS()
{
  fprintf(stderr, "\nExiting IQLS due to error(s) listed above.\n\n");
  exit(1);
}


int check_paramset(ParamSet &ps)
{
  int status = 1;
  
  if(strlen(ps.phenofile)>0)
  {
    if(!fopen(ps.phenofile, "r"))
    {
      fprintf(stderr, "Error: cannot open phenotype data file \"%s\".\n", ps.phenofile);
      status = -1;
    }
  }

  if(strlen(ps.genofile)>0)
  {
    if(!fopen(ps.genofile, "r"))
    {
      fprintf(stderr, "Error: cannot open genotype data file \"%s\".\n", ps.genofile);
      status = -1;
    }
  }
  
  if(strlen(ps.ibdfile)>0)
  {
    if(!fopen(ps.ibdfile, "r"))
    {
      fprintf(stderr, "Error: cannot open IBD coefficients file \"%s\".\n", ps.ibdfile);
      status = -1;
    }
  }

  if(strlen(ps.parameterfile)>0)
  {
    if(!fopen(ps.parameterfile, "r"))
    {
      fprintf(stderr, "Error: cannot open parameter file \"%s\".\n", ps.parameterfile);
      status = -1;
    }
  }
  
  return status;
}


void invoke_IQLS(ParamSet &ps)
{
  PedData ped;

  FILE *fp = fopen("IQLStest.out", "w");
  if(!fp)
  {
    fprintf(stderr, "Error: cannot open output file \"IQLStest.out\" for writing.\n");
    kill_IQLS();
  }

  FILE *fpp = fopen("IQLStest.fdf.top", "w");
  if(!fpp)
  {
	fprintf(stderr, "Error: cannot open significance file \"IQLStest.fdf.top\" for writing.\n");
    kill_IQLS();
  }

  FILE *fppp = fopen("IQLStest.1df.top", "w");
  if(!fppp)
  {
    fprintf(stderr, "Error: cannot open output file \"IQLStest.1df.top\" for writing.\n");
    kill_IQLS();
  }
  
  ped.set_output(fp,fpp,fppp);
  
  if(ped.run(ps)<0)
    kill_IQLS();
  fprintf(stderr, "\nIQLS is completed\n\n");
}


int main(int argc, char **argv)
{
  ParamSet ps;

  // input files
  for(int i=1;i<argc;i++)
  {
    if(strcmp(argv[i],"-pheno")==0 || strcmp(argv[i],"-PHENO")==0)
    {
      strcpy(ps.phenofile,argv[++i]);
      continue;
    }
    if(strcmp(argv[i],"-geno")==0 || strcmp(argv[i],"-GENO")==0)
    {
      strcpy(ps.genofile,argv[++i]);
      continue;
    }
    if(strcmp(argv[i],"-ibd")==0 || strcmp(argv[i],"-IBD")==0)
    {
      strcpy(ps.ibdfile,argv[++i]);
      continue;
    }
    if(strcmp(argv[i],"-r")==0 || strcmp(argv[i],"-R")==0)
    {
      strcpy(ps.parameterfile,argv[++i]);
      continue;
    }

    if(strcmp(argv[i],"-b")==0 || strcmp(argv[i],"-B")==0)
    {
      ps.begin = atoi(argv[++i]);
      continue;
    }
    if(strcmp(argv[i],"-e")==0 || strcmp(argv[i],"-E")==0)
    {
      ps.end = atoi(argv[++i]);
      continue;
    }

    if(strcmp(argv[i],"-op")==0 || strcmp(argv[i],"-OP")==0)
    {
      ps.outoption = (argv[++i])[0];
      continue;
    }
    if(strcmp(argv[i],"-thresh")==0 || strcmp(argv[i],"-THRESH")==0)
    {
      ps.full_thresh = atoi(argv[++i]);
      continue;
    }
    
    fprintf(stderr, "Error: Unknown command options \"%s\"\n", argv[i]);
    kill_IQLS();
  }

  int status = check_paramset(ps);

  if(status<0)
    kill_IQLS();

  invoke_IQLS(ps);

  return 1;
}
